/*
 * Brzozowski.cpp
 *
 *  Created on: 11. 1. 2014
 *      Author: tomas
 */

#include "Brzozowski.h"

using namespace alib;
using namespace automaton;
using namespace regexp;

namespace conversions
{

Brzozowski::Brzozowski( const RegExp & re ) : AbstractREtoFAConverter( re )
{

}

Brzozowski::~Brzozowski( void )
{

}

FSM Brzozowski::convert( void )
{
    RegExpOptimize opt;

    RegExp V = opt.optimize( m_re );
    set<RegExpSymbol> alphabet = RegExpAlphabet::getSymbols( m_re );

    set<RegExp> Q = { V };
    deque<set<RegExp>> Qi;

    Qi.push_back( set<RegExp>( ) );
    Qi.at( 0 ).insert( V );

    int i = 1;
    while( ! Qi.at( i - 1 ).empty( ) )
    {
        Qi.push_back( set<RegExp>( ) ); // initialize set Q_i

        for( const auto & regexp : Qi.at( i - 1 ) )
        {
            RegExpDerivation deriv( regexp );

            for( const auto & a : alphabet )
            {
                RegExpElement* dSymbol = new RegExpSymbol( a.getSymbol( ) );
                RegExp derived = deriv.derivation( list<RegExpElement*>( 1, dSymbol ) );
                derived = opt.optimize( derived );

                // no "trash state" (though algorithm probably considers it)
                if( ! derived.isEmpty( ) )
                {
                    if( ! isInSet( derived, Q ) ) // if this state has already been found, do not add
                        Qi.at( i ).insert( derived );

                    m_transitions.insert( Transition( regexp, Symbol( a.getSymbol( ) ), derived ) );
                }

                delete dSymbol;
            }
        }

        Q.insert( Qi.at( i ).begin( ), Qi.at( i ).end( ) );
        i += 1;
    }

    // ------------------------------------------------------------------------

    StateBuilder builder( Q );

    for( const auto & r : Q )
        m_fsm.addState( builder.getState( r ) );

    for( const auto & a : alphabet )
        m_fsm.addInputSymbol( a.getSymbol( ) );

    for( const auto & t : m_transitions )
        m_fsm.addTransition( builder.getState( t.m_from ), t.m_symbol, builder.getState( t.m_to ) );

    m_fsm.addInitialState( builder.getState( V ) );

    for( const auto & U : Q )
        if( U.containsEmptyString( ) )
            m_fsm.addFinalState( builder.getState( U ) );

    return m_fsm;
}

// ----------------------------------------------------------------------------

Brzozowski::Transition::Transition( const RegExp & from, const Symbol & symb, const RegExp & to )
    : m_from( from ), m_to( to ), m_symbol( symb )
{

}

bool Brzozowski::Transition::operator<( const Transition & x ) const
{
    if( m_from != x.m_from )
        return m_from < x.m_from;
    else if( m_symbol != x.m_symbol )
        return m_symbol < x.m_symbol;
    else
        return m_to < x.m_to;
}

// ----------------------------------------------------------------------------

Brzozowski::StateBuilder::StateBuilder( const set<RegExp> & Q )
{
    m_stateId = 0;

    for( const auto & regexp : Q )
        m_states.insert( pair<RegExp, State>( regexp, State( createNewName( ) ) ) );
}

const State & Brzozowski::StateBuilder::getState( const RegExp & re ) const
{
    // map::find() does not work! see gitlab issue #12
    for( const auto & kv : m_states )
        if( kv.first == re )
            return kv.second;

    throw AlibException( "Brzozowski::StateBuilder - Regular expression not found!" );
}

string Brzozowski::StateBuilder::createNewName( void )
{
    // http://en.wikipedia.org/wiki/Hexavigesimal

    unsigned int n = m_stateId ++;
    string name;
    do
    {
        unsigned int remainder = n % 26;
        name += ( char )( remainder + 'A' );
        n = (n - remainder) / 26;
    } while (n > 0);

    return string( name.rbegin( ), name.rend( ) );
}

} /* namespace conversions */
