/*
 * GlushkovTraversal.cpp
 *
 *  Created on: 13. 3. 2014
 *      Author: tomas
 */

#include "GlushkovTraversal.h"

using namespace alib;
using namespace regexp;
using namespace std;

namespace conversions
{

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const RegExp & re )
{
    return getLeftmostSymbolsInTree( re.getRegExp( ) );

}
set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const RegExp & re )
{
    return getRightmostSymbolsInTree( re.getRegExp( ) );
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const RegExp & re )
{
    return getNeighbours( re.getRegExp( ) );
}

// -----------------------------------------------------------------------------

GlushkovTraversal::Neighbours::Neighbours( const RegExpSymbol * first, const RegExpSymbol * second ) :
        m_first( first ),
        m_second( second )
{

}

bool GlushkovTraversal::Neighbours::operator<( const Neighbours & x ) const
{
    if( m_first != x.m_first )
        return m_first < x.m_first;
    else
        return m_second < x.m_second;
}

// -----------------------------------------------------------------------------

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const RegExpElement * node )
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( symbol )
        return getLeftmostSymbolsInTree( symbol );
    else if( alternation )
        return getLeftmostSymbolsInTree( alternation );
    else if( concatenation )
        return getLeftmostSymbolsInTree( concatenation );
    else if( iteration )
        return getLeftmostSymbolsInTree( iteration );
    else if( eps )
        return getLeftmostSymbolsInTree( eps );
    else if( empty )
        return getLeftmostSymbolsInTree( empty );

     throw AlibException( "GlushkovTraversal::getLeftmostSymbolsInTree - invalid RegExpElement node" );
}

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const Alternation * node )
{
    set<const RegExpSymbol*> ret;

    for( const auto & e : node->getElements( ) )
    {
        const set<const RegExpSymbol*> tmp = getLeftmostSymbolsInTree( e );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const Concatenation * node )
{
    set<const RegExpSymbol*> ret;

    for( const auto & e : node->getElements( ) )
    {
        set<const RegExpSymbol*> tmp = getLeftmostSymbolsInTree( e );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! e->containsEmptyString( ) ) // If this subtree can be epsilon, then we need to add next subtree also
            break;
    }

    return ret;
}

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const Iteration * node )
{
    return getLeftmostSymbolsInTree( node->getElement( ) );
}

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const RegExpSymbol * node )
{
    return set<const RegExpSymbol*> { node };
}

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const RegExpEpsilon * node )
{
    return set<const RegExpSymbol*>( );
}

set<const RegExpSymbol*> GlushkovTraversal::getLeftmostSymbolsInTree( const RegExpEmpty * node )
{
    return set<const RegExpSymbol*>( );
}
// ----------------------------------------------------------------------------

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const RegExpElement * node )
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( symbol )
        return getRightmostSymbolsInTree( symbol );
    else if( alternation )
        return getRightmostSymbolsInTree( alternation );
    else if( concatenation )
        return getRightmostSymbolsInTree( concatenation );
    else if( iteration )
        return getRightmostSymbolsInTree( iteration );
    else if( eps )
        return getRightmostSymbolsInTree( eps );
    else if( empty )
        return getRightmostSymbolsInTree( empty );

     throw AlibException( "GlushkovTraversal::getRightmostSymbolsInTree - invalid RegExpElement node" );
}

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const Alternation * node )
{
    set<const RegExpSymbol*> ret;

    for( const auto & e : node->getElements( ) )
    {
        set<const RegExpSymbol*> tmp = getRightmostSymbolsInTree( e );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const Concatenation * node )
{
    set<const RegExpSymbol*> ret;

    for( auto it = node->getElements( ).rbegin( ); it != node->getElements( ).rend( ) ; it ++ )
    {
        set<const RegExpSymbol*> tmp = getRightmostSymbolsInTree( *it );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! ( * it )->containsEmptyString( ) )
            break;
    }

    return ret;
}

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const Iteration * node )
{
    return getRightmostSymbolsInTree( node->getElement( ) );
}

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const RegExpSymbol * node )
{
    return set<const RegExpSymbol*> { node };
}

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const RegExpEpsilon * node )
{
    return set<const RegExpSymbol*>( );
}

set<const RegExpSymbol*> GlushkovTraversal::getRightmostSymbolsInTree( const RegExpEmpty * node )
{
    return set<const RegExpSymbol*>( );
}

// ----------------------------------------------------------------------------

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const RegExpElement * node )
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( symbol )
        return getNeighbours( symbol );
    else if( alternation )
        return getNeighbours( alternation );
    else if( concatenation )
        return getNeighbours( concatenation );
    else if( iteration )
        return getNeighbours( iteration );
    else if( eps )
        return getNeighbours( eps );
    else if( empty )
        return getNeighbours( empty );

     throw AlibException( "GlushkovTraversal::getNeighbours - unknown RegExpElement* " );
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const Alternation * node )
{
    set<Neighbours> n;
    for( const auto & e : node->getElements( ) )
    {
        set<Neighbours> tmp = getNeighbours( e );
        n.insert( tmp.begin( ), tmp.end( ) );
    }

    return n;
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const Concatenation * node )
{
    set<Neighbours> n;
    for( const auto & e : node->getElements( ) )
    {
        set<Neighbours> tmp = getNeighbours( e );
        n.insert( tmp.begin( ), tmp.end( ) );
    }

    for( auto e = node->getElements( ).begin( ); e != node->getElements( ).end( ); e ++ )
    {
        auto f = e;
        if( f == node->getElements( ).end( ) )
            continue;

        for( f++ ; f != node->getElements( ).end( ); f ++ )
        {
            for( const auto & x : getRightmostSymbolsInTree( * e ) )
                for( const auto & y : getLeftmostSymbolsInTree( * f ) )
                    n.insert( Neighbours( x, y ) );

            if( ! ( * f )->containsEmptyString( ) )
                break;
        }
    }

    return n;
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const Iteration * node )
{
    set<Neighbours> n;
    set<Neighbours> tmp = getNeighbours( node->getElement( ) );
    n.insert( tmp.begin( ), tmp.end( ) );

    for( const auto & x : getRightmostSymbolsInTree( node->getElement( ) ) )
        for( const auto & y : getLeftmostSymbolsInTree( node->getElement( ) ) )
            n.insert( Neighbours( x, y ) );

    return n;
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const RegExpSymbol * node )
{
    return set<Neighbours>( );
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const RegExpEpsilon * node )
{
    return set<Neighbours>( );
}

set<GlushkovTraversal::Neighbours> GlushkovTraversal::getNeighbours( const RegExpEmpty * node )
{
    return set<Neighbours>( );
}

} /* namespace conversions */
